@@ -98,13 +98,10 @@ typedef struct {
 GType nice_stream_get_type (void);
 
 NiceStream *
-nice_stream_new (guint n_components, NiceAgent *agent);
+nice_stream_new (guint stream_id, guint n_components, NiceAgent *agent);
 
 void
-nice_stream_close (NiceStream *stream);
-
-gboolean
-nice_stream_all_components_ready (NiceStream *stream);
+nice_stream_close (NiceAgent *agent, NiceStream *stream);
 
 NiceComponent *
 nice_stream_find_component_by_id (NiceStream *stream, guint id);
